import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import processing.controlP5.ControlP5;
import processing.controlP5.Slider;
import processing.controlP5.Toggle;
import processing.core.*;
import processing.video.*;

public class FastBlobTracking extends PApplet {
	
	private static final String PRESET_PATH = "presets.txt";

	private BlobedCapture cam;
	private PGraphics result;
	private int prevtime;
	private int deltatime;

	private ControlP5 ui;
	private Slider ui_bgtol;
	private Slider ui_frstack;
	private Slider ui_bmin;
	private Slider ui_bmax;
	private Slider ui_bthr;
	private Slider ui_bcon;
	private Toggle ui_save;
	private Toggle ui_reload;
	private Toggle ui_reset;
	
	private boolean saverequest;
	
	public int bg_tolerance;
	public int frame_stack_num;
	public float blob_min_size;
	public float blob_max_size;
	public float blob_threshold;
	public float blob_connection_radius;
	public boolean save_presets;
	public boolean reload_presets;
	public boolean reset_presets;

	public FastBlobTracking() {
		
		super();
		prevtime = 0;
		deltatime = 0;
		saverequest = false;
		save_presets = false;
		reload_presets = false;
		reset_presets = false;
		
		bg_tolerance = 0;
		frame_stack_num = 0;
		blob_min_size = 0;
		blob_max_size = 0;
		blob_threshold = 0;
		blob_connection_radius = 0;
		
	}

	public void settings() {
		
		size( 1280, 480, P3D );
	
	}

	@SuppressWarnings("deprecation")
	public void setup() {
	
		// ps eye on linux (/dev/video0 is the integrated webcam on my computer)
//		cam = new BlobedCapture( this, 320, 240, "/dev/video1", 100 );
		
		// integrated cam, when ps eye is not connected
		cam = new BlobedCapture( this, 320, 240, "/dev/video0", 30 );
		cam.start();
		
		result = createGraphics( 640, 480, P2D );
		result.beginDraw();
		result.background( 5 );
		result.endDraw();

		int uix = 970;
		int uiy = 10;
		
		ui = new ControlP5( this );
		ui.addLabel( "BACKGROUND SUBSTRACTION", uix, uiy ); 
		uiy += 15; 
		ui_bgtol = ui.addSlider( "bg_tolerance", 0, 255, uix, uiy, 100, 10 );
		uiy += 25;
		ui.addLabel( "STACK", uix, uiy );
		uiy += 15; 
		ui_frstack = ui.addSlider( "frame_stack_num", 1, 200, uix, uiy, 100, 10 );
		uiy += 25;
		ui.addLabel( "BLOBS", uix, uiy );
		uiy += 15;
		ui_bmin = ui.addSlider( "blob_min_size", 0, 1, uix, uiy, 100, 10 );
		uiy += 15;
		ui_bmax = ui.addSlider( "blob_max_size", 0, 1, uix, uiy, 100, 10 );
		uiy += 15;
		ui_bthr = ui.addSlider( "blob_threshold", 0, 1, uix, uiy, 100, 10 );
		uiy += 15;
		ui_bcon = ui.addSlider( "blob_connection_radius", 0, 1, uix, uiy, 100, 10 );
		uiy += 25;
		ui.addLabel( "PRESETS", uix, uiy );
		uiy += 15;
		ui_save = ui.addToggle( "save_presets", false, uix, uiy, 20, 20 );
		ui_reload = ui.addToggle( "reload_presets", false, uix + 90, uiy, 20, 20 );
		ui_reset = ui.addToggle( "reset_presets", false, uix + 180, uiy, 20, 20 );
		
		loadPresets();
		
	}
	
	public void captureEvent( Capture cam ) {
		((BlobedCapture) cam).configure( bg_tolerance, frame_stack_num, blob_min_size, blob_max_size, blob_threshold, blob_connection_radius );
		cam.read();
		int ms = millis();
		if ( prevtime != 0 ) {
			deltatime = ms - prevtime;
		}
		prevtime = ms;
	
	}

	public void draw() {
		
		while( cam.isLocked() ) {
			try {
				Thread.sleep( 1 );
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		
		background( 5 );
		
		image( cam.getImg(), 0, 0 );
		image( cam.getDiff(), 0, cam.height );

		// display
		pushMatrix();
		translate( cam.width, 0 );
		rectMode( CENTER );
		noStroke();
		String frameinfo = "";
		ArrayList< float[] > blobframes = cam.cloneFrames();
		int i = 0;
		for ( float[] frame : blobframes ) {
			int w = 75 + (int) ( 180 * ( i * 1.f / cam.getFramestacknum() ) );
			fill( w, 50 );
			for ( int f = 0; f < frame.length; f += 5 ) {
				rect( 
					frame[ f ] * 640, 
					frame[ f + 1 ] * 480, 
					frame[ f + 2 ] * 640, 
					frame[ f + 3 ] * 480 );
			}
			frameinfo += "" + i + ": " + frame.length  +"\n";
			i++;
		}
		stroke( 255, 0, 0 );
		float[] prevframe = null;
		for ( float[] frame : blobframes ) {
			for ( int f = 0; f < frame.length; f += 5 ) {
				if ( frame[ f + 4 ] == -1 || prevframe == null ) continue;
				line( 
					frame[ f ] * 640, 
					frame[ f + 1 ] * 480, 
					prevframe[ (int) frame[ f + 4 ] ] * 640, 
					prevframe[ (int) frame[ f + 4 ] + 1 ] * 480 );
			}
			prevframe = frame;
		}
		popMatrix();
		
		fill( 255 );
		text( "display: " + frameRate, 10, 25 );
		text( "camera: " + deltatime + " millis", 10, 40 );
		text( "process: " + cam.getProcesstime() + " millis", 10, 55 );
		text( "frames info:\n" + frameinfo, 10, 70 );
		
		if ( saverequest ) {
			saveFrame( "screenshot_" + frameCount + ".jpg" );
			saverequest = false;
		}
		if ( save_presets ) {
			savePresets();
			save_presets = false;
			ui_save.changeValue( 0 );
		}
		if ( reload_presets ) {
			loadPresets();
			reload_presets = false;
			ui_reload.changeValue( 0 );
		}
		if ( reset_presets ) {
			resetPresets();
			reset_presets = false;
			ui_reset.changeValue( 0 );
		}
		
	}
	
	private void savePresets() {
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter( PRESET_PATH ));
            bw.write("bg_tolerance=" + bg_tolerance + "\n");
            bw.write("frame_stack_num=" + frame_stack_num + "\n");
            bw.write("blob_min_size=" + blob_min_size + "\n");
            bw.write("blob_max_size=" + blob_max_size + "\n");
            bw.write("blob_threshold=" + blob_threshold + "\n");
            bw.write("blob_connection_radius=" + blob_connection_radius + "\n");
            bw.close();
        } catch (IOException e) {
        	e.printStackTrace();
        }	
	}
	
	private void syncUI() {
		
		ui_bgtol.setValue( bg_tolerance );
		ui_frstack.setValue( frame_stack_num );
		ui_bmin.setValue( blob_min_size );
		ui_bmax.setValue( blob_max_size );
		ui_bthr.setValue( blob_threshold );
		ui_bcon.setValue( blob_connection_radius );
		
	}
	
	private void resetPresets() {

		bg_tolerance = BlobedCapture.BACKGROUND_TOLERANCE_DEFAULT;
		frame_stack_num = BlobedCapture.FRAMES_STACK_NUM_DEFAULT;
		blob_min_size = BlobedCapture.BLOB_MINIMUM_SIZE_DEFAULT;
		blob_max_size = BlobedCapture.BLOB_MAXIMUM_SIZE_DEFAULT;
		blob_threshold = BlobedCapture.BLOB_THRESHOLD_DEFAULT;
		blob_connection_radius = BlobedCapture.CONNECTION_RADIUS_DEFAULT;
		
		syncUI();
		
	}
	
	private void loadPresets() {
		
		// setting default values, just in case
		resetPresets();
		
		try {
			BufferedReader br = new BufferedReader(new FileReader( PRESET_PATH ));
			String line = br.readLine();
			while (line != null) {
				line = line.replaceAll("\\s","");
				String[] words = line.split( "=" );
				if ( words.length == 2 ) {
					if ( words[ 0 ].equals( "bg_tolerance" ) ) {
						bg_tolerance = Integer.parseInt( words[ 1 ] );
					} else if ( words[ 0 ].equals( "frame_stack_num" ) ) {
						frame_stack_num = Integer.parseInt( words[ 1 ] );
					} else if ( words[ 0 ].equals( "blob_min_size" ) ) {
						blob_min_size = Float.parseFloat( words[ 1 ] );
					} else if ( words[ 0 ].equals( "blob_max_size" ) ) {
						blob_max_size = Float.parseFloat( words[ 1 ] );
					} else if ( words[ 0 ].equals( "blob_threshold" ) ) {
						blob_threshold = Float.parseFloat( words[ 1 ] );
					} else if ( words[ 0 ].equals( "blob_connection_radius" ) ) {
						blob_connection_radius = Float.parseFloat( words[ 1 ] );
					} else {
						println( words[ 0 ]  + "?" );
					}
				} else {
					println( line + "?" );
				}
				line = br.readLine();
			}
			br.close();
        } catch (IOException e) {
        	e.printStackTrace();
        }
        
        syncUI();
        
	}
	
	public void keyPressed() {
		
		if ( key == 's' ) {
			saverequest = true;
		} else {
			// reactivate to print the key pressed
			println( (char) key + " / " + keyCode );
		}
		
	}
	
	public static void main(String args[]) {
		
		PApplet.main(new String[] { "FastBlobTracking" });
	
	}

}
